/*
 * A simple DECnet bridge program
 * (c) 2003, 2005 by Johnny Billquist
 *
 *
 * Original work of Johnny Billquist (bqt@update.uu.se)
 *
 *
 * Subsequent updates:
 *
 *      V0.7-002               Federico Di Dio                27-AUG-2007
 *           Fixed minor portability issues, should work on OpenBSD.
 *
 *      V0.7-001               Fabio Battaglia                19-AUG-2007
 *           Added multiple local ethernet support. Now we can listen to and
 *           (if desired) forward between local interfaces. Just add into the
 *           configuration file more than one ethernet line in the [bridge] and
 *           [forward] sections as needed. To forward between any two local
 *           ethernet interfaces enter just one of them into the [forward] sect.
 *           
 *      V0.6-003               Gerardo Cacciari               17-AUG-2007
 *           New -b option to force binding to a specified IP address instead
 *           of INADDR_ANY (0.0.0.0). Genericized BSD code using a common ifdef
 *           __BSD__ clause. These changes along with those introduced by
 *           V0.6-1 should cover every routing problem, we hope.
 *
 *      V0.6-002               Gerardo Cacciari               14-AUG-2007
 *           Some minor bug fixes and editing changes. Function dump_data()
 *           splitted into dump_config() and dump_table(). The latter is called
 *           at every DNS lookup cycle when in debug mode. No more special
 *           handling for some signals (SIGHUP and SIGUSR1).
 *
 *      V0.6-001               Gerardo Cacciari               06-AUG-2007
 *           Made some changes coming from Johnny Billquist's bridge v. 2.0
 *           New feature that allows incoming frames from any source UDP port
 *           This new feature is controlled by a new .conf specific section.
 *           
 *      V0.5-003               Fabio Battaglia                02-JUN-2007
 *           New genericized daemonization code, some changes to BSD  
 *           conditional includes, some minor cosmetic changes.
 *
 *      V0.5-002               Gerardo Cacciari               04-FEB-2007
 *           New versioning scheme and new change log in the program header
 *           A little banner is displayed upon program startup.
 *
 *      V0.5-001               Gerardo Cacciari               02-FEB-2007
 *           Added selective forward and moved passive settings to a specific
 *           configuration file section (instead of '~' as the first hostname
 *           character in the configuration file). Again some more code
 *           cleanup (mainly debug messages).
 *
 *      V0.4-001               Gerardo Cacciari               24-DEC-2006
 *           Added more debug controls and sanity checks.
 *           Fixed a bug in DNS lookups.
 *
 *      V0.3-002               Fabio Battaglia                21-DEC-2006
 *           Some more debug controls and messages.
 * 
 *      V0.3-001               Gerardo Cacciari               21-DEC-2006
 *           Time-controlled DNS lookups (instead of configuration file reload)
 *           and some code cleanup.
 *
 *      V0.2-001               Fabio Battaglia                19-DEC-2006
 *           Command line parsing and daemonization.
 *
 *      V0.1-002               Fabio Battaglia                18-DEC-2006
 *           Configuration file reloadable every N seconds.
 *
 *      V0.1-001               Gerardo Cacciari               18-DEC-2006
 *           Default forwarding disabled (for mesh-structured networks).
 *           First changes over the original bridge from Johnny Billquist.
 *
 *
 * http://decnet.ipv7.net  -  #retrocomputing on irc.azzurra.org
 *
 * Gerardo Cacciari (gerry77@mail.com)
 * Fabio Battaglia (nightshade@email.it - http://belzebub.servebeer.com)
 * Federico Di Dio (fdd@paolyno.ath.cx)
 *
 */

#define BRIDGE_VERSION 0
#define BRIDGE_RELEASE 7
#define BRIDGE_EDITLVL 2

#define DEFAULT_PORT 4711
#define DEFAULT_CONFIG "/etc/decnet-bridge.conf"

#define MAX_HOST 16

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#   define __BSD__
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <time.h>
#include <sys/ioctl.h>
#ifdef __BSD__
  #include <net/bpf.h>
#else
  #include <pcap-bpf.h>
#endif
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <pcap.h>
#include <netdb.h>
#include <signal.h>
#include <string.h>

static unsigned long refresh_interval = 0;
static unsigned int debug = 0;
static char* config_name = DEFAULT_CONFIG;

/*
   Throttling control:
   THROTTLETIME - (uS)
                  If packets come closer in time than this, they are
                  a base for perhaps considering throttling.
   THROTTLEPKT  - (#)
                  The number of packets in sequence that fulfill
                  THROTTLETIME that means throttling will kick in.
   THROTTLEDELAY - (uS)
                  The delay to insert when throttling is active.
*/

#define THROTTLETIME 5000
#define THROTTLEPKT 4
#define THROTTLEDELAY 10

#define PASSIVE_TMO 120000000L

#define THROTTLEMASK ((1 << THROTTLEPKT) - 1)

#define ETHERTYPE_DECNET 0x6003
#define ETHERTYPE_LAT 0x6004
#define ETHERTYPE_MOPDL 0x6001
#define ETHERTYPE_MOPRC 0x6002

#ifndef MAX
  #define MAX(a,b) (((a)>(b))?(a):(b))
#endif

/*
   This is a very simple and small program for bpf that just
   filters out anything by any protocol that we *know* we're
   not interested in.
*/

struct bpf_insn insns[] = {
  BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12),
  BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETHERTYPE_MOPRC, 4, 0),
  BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETHERTYPE_MOPDL, 3, 0),
  BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETHERTYPE_LAT, 2, 0),
  BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, ETHERTYPE_DECNET, 1, 0),
  BPF_STMT(BPF_RET+BPF_K, 0),
  BPF_STMT(BPF_RET+BPF_K, 1518),
};

/* 
   The structures and other global data we keep info in.
   It would perhaps be nice if we could reload this, and
   in case of failure keep the old stuff, but for now we
   don't care that much...

   The data structures we have are the port, which describe
   a source/destination for data. It also holds info about which
   kind of traffic should be forwarded to this site. It is also
   used to filter incoming packets. If we don't send something, we
   don't want it from that side either.
   We have the host table, which is a hash table for all known
   destinations, so that we can optimize the traffic a bit.

   When data arrives, we filter, process and send it out again.
*/

typedef enum {DECnet, LAT} pkttyp;
#define MAXTYP 2

struct BRIDGE {
  char name[40];
  char host[80];
  in_addr_t addr;
  int port;
  int passive;
  int fd;
  int types[MAXTYP];
  char last[8][14];
  int lastptr;
  int rcount;
  int tcount;
  int xcount;
  struct timeval lasttime;
  int throttle;
  int throttlecount;
  struct timeval lastrcv;
  int forward;
  int anyport;
  pcap_t *pcap;
};

struct DATA {
  int source;
  pkttyp type;
  int len;
  const char *data;
};

struct HOST {
  struct HOST *next;
  unsigned char mac[6];
  int bridge;
  struct timeval lastseen;
};

#define HOST_HASH 65536

struct HOST *hosts[HOST_HASH];
struct BRIDGE bridge[MAX_HOST];
int bcnt = 0;
int sd;

/* Here comes the code... */

int lookup(struct sockaddr_in *sa)
{
  int i;

  if(debug) fprintf(stdout,"lookup(): packet from %s port %d ",inet_ntoa(sa->sin_addr),ntohs(sa->sin_port));
  for (i=0; i<bcnt; i++) {
    if ((bridge[i].addr == sa->sin_addr.s_addr) &&
        ((bridge[i].port == sa->sin_port) || (bridge[i].anyport != 0))) {
      if(debug) fprintf(stdout,"accepted for bridge %d (%s)\n",i,bridge[i].name);
      return i;
    }
  }
  if(debug) fprintf(stdout,"refused\n");
  return -1;
}


int lookup_bridge(char *newbridge)
{
  int i, bridgelen = strlen(newbridge);
  for (i=0; i<bcnt; i++)
    if ((strcmp(newbridge,bridge[i].name) == 0) && 
        (bridgelen == strlen(bridge[i].name))) return i; /* bqt 2.0 */
  return -1;
}


void add_bridge(char *name, char *dst)
{
  struct hostent *he;
  char rhost[40];
  int port = 0;
  int i,found = 0;
  struct in_addr addr = {0};
  char *p;

  if (bcnt < MAX_HOST) {
    bzero(&bridge[bcnt],sizeof(struct BRIDGE));
    strncpy(bridge[bcnt].name,name,(size_t)40);
    p = index(dst,':');
    if (p == NULL) {        /* Assume local descriptor */
      struct bpf_program pgm;
      char ebuf[PCAP_ERRBUF_SIZE];

      if ((bridge[bcnt].pcap = pcap_open_live(dst, 1518, 1, 0, ebuf)) == 0) {
        fprintf(stderr,"ERROR: unable to open local device %s\n", dst);
        /* pcap_perror(bridge[bcnt].pcap, ebuf); */
        exit(1);
      }

      pgm.bf_len=sizeof(insns)/sizeof(struct bpf_insn);
      pgm.bf_insns=insns;
      if (pcap_setfilter(bridge[bcnt].pcap, &pgm) == -1) {
        pcap_perror(bridge[bcnt].pcap, "loading filter program");
        exit(1);
      }

      strncpy(bridge[bcnt].host,dst,(size_t)80);
      bridge[bcnt].addr = 0;
      bridge[bcnt].port = 0;
      bridge[bcnt].fd = pcap_fileno(bridge[bcnt].pcap);

#ifdef __BSD__
      i = 1;
      if (ioctl(bridge[bcnt].fd,BIOCIMMEDIATE,&i) == -1) {
        perror("BIOCIMMEDIATE");
        exit(1);
      }
      if (ioctl(bridge[bcnt].fd,BIOCSHDRCMPLT,&i) == -1) {
        perror("BIOCSHDRCMPLT");
        exit(1);
      }
#endif

      found = -1;

    } else {

      *p = ' ';
      sscanf(dst,"%s %d", rhost, &port);
      strncpy(bridge[bcnt].host,rhost,(size_t)80);
      bridge[bcnt].fd = sd;

      bridge[bcnt].port = htons(port);
      if (bridge[bcnt].port < 1025 || bridge[bcnt].port > 32767) {
        fprintf(stderr,"ERROR: port number %d for bridge %s is out of range (1025-32767)\n", bridge[bcnt].port, name);
        exit(1);
      }

      if ((he = gethostbyname(rhost)) != NULL) {
        addr.s_addr = *(uint32_t *)he->h_addr;
        found = -1;
      } else {
        found = inet_aton(rhost,&addr);
      }
      if (found) {
        bridge[bcnt].addr = addr.s_addr;
      }
    }
    if (found) {
      for (i=0; i<MAXTYP; i++) bridge[bcnt].types[i] = 0;

      bridge[bcnt].rcount = 0;
      bridge[bcnt].tcount = 0;
      bridge[bcnt].passive = 0; 

      bcnt++;
      if(debug)fprintf(stdout,"Adding bridge %s %s:%d\n", name, inet_ntoa(addr), port);
 
    } else {
      fprintf(stderr,"ERROR: invalid or not resolved address %s for bridge %s\n", rhost, name);
    }
  } else {
    fprintf(stderr,"ERROR: bridge table full. Not adding %s (%s)\n", name, dst);
  }
}


int add_service(char *newbridge, pkttyp type, char *name)
{
  int i;
  if(debug)fprintf(stdout,"Adding %s service %s.\n", name, newbridge);
  if ((i = lookup_bridge(newbridge)) >= 0) {
    if (bridge[i].types[type]++ > 0) {
      if(debug)fprintf(stdout,"%s service %s added multiple times.\n", name, newbridge);
    }
    return 1;
  }
  return 0;
}


int set_passive(char *newbridge)
{
  int i;
  if(debug)fprintf(stdout,"Enabling passive option for bridge %s.\n", newbridge);
  if ((i = lookup_bridge(newbridge)) >= 0) {
    if (bridge[i].passive++ > 0) {
      if(debug)fprintf(stdout,"Passive option for bridge %s enabled multiple times.\n", newbridge);
    } 
    return 1;
  }
  return 0;
}


int set_forward(char *newbridge)
{
  int i;
  if(debug)fprintf(stdout,"Enabling forward option for bridge %s.\n", newbridge);
  if ((i = lookup_bridge(newbridge)) >= 0) {
    if (bridge[i].forward++ > 0) {
      if(debug)fprintf(stdout,"Forward option for bridge %s enabled multiple times.\n", newbridge);
    }
    return 1;
  }
  return 0;
}


int set_anyport(char *newbridge)
{
  int i;
  if(debug)fprintf(stdout,"Enabling anyport option for bridge %s.\n", newbridge);
  if ((i = lookup_bridge(newbridge)) >= 0) {
    if (bridge[i].anyport++ > 0) {
      if(debug)fprintf(stdout,"Anyport option for bridge %s enabled multiple times.\n", newbridge);
    }
    return 1;
  }
  return 0;
}


void read_conf()
{
  FILE *f;
  int mode = 0;
  int line;
  char buf[80];
  char buf1[40],buf2[40];
  int i;

  if(debug) fprintf(stdout,"DECnet bridge reading config file\n");

  if ((f = fopen(config_name,"r")) == NULL) {
    fprintf(stderr,"ERROR: unable to open configuration file %s\n",config_name);
    exit(1);
  }

  for (i=0; i<bcnt; i++) {
    if (bridge[i].fd != sd) close(bridge[i].fd);
  }
  bcnt = 0;

  for (i=0; i<HOST_HASH; i++) {
    struct HOST *h, *n;
    h = hosts[i];
    hosts[i] = NULL;
    while(h) {
      n = h->next;
      free(h);
      h = n;
    }
  }

  line = 0;
  while (!feof(f)) {
    if (fgets(buf,80,f) == NULL) continue; /* bqt 2.0 */
    buf[strlen(buf)-1] = 0;
    line++;
    if((strlen(buf) > 2) && (buf[0] != '!')) {
      if(buf[0]=='[') {
        mode = -1;
        if(strcmp(buf,"[bridge]") == 0) mode = 0;
        if(strcmp(buf,"[decnet]") == 0) mode = 1;
        if(strcmp(buf,"[lat]") == 0) mode = 2;
        if(strcmp(buf,"[passive]") == 0) mode = 3;
        if(strcmp(buf,"[forward]") == 0) mode = 4;
        if(strcmp(buf,"[anyport]") == 0) mode = 5;
        if(mode < 0) {
          fprintf(stderr,"ERROR: bad configuration section %s at line %d\n", buf, line);
          exit(1);
        }
      } else {
        switch (mode) {
        case 0:
          if (sscanf(buf, "%s %s", buf1, buf2) == 2) {
            add_bridge(buf1,buf2);
          } else {
            fprintf(stderr,"ERROR: remote bridge definition %s at line %d is not correct\n", buf, line);
            exit(1);
          }
          break;
        case 1:
          if (!add_service(buf,DECnet,"DECnet")) {
            fprintf(stderr,"ERROR: DECnet service %s at line %d does not match any bridge definition\n", buf, line);
            exit(1);
          } 
          break;
        case 2:
          if (!add_service(buf,LAT,"LAT")) {
            fprintf(stderr,"ERROR: LAT service %s at line %d does not match any bridge definition\n", buf, line);
            exit(1);
          }
          break;
        case 3:
          if (!set_passive(buf)) {
            fprintf(stderr,"ERROR: passive option %s at line %d does not match any bridge definition\n", buf, line);
            exit(1);
          }
          break;
        case 4:
          if (!set_forward(buf)) {
            fprintf(stderr,"ERROR: forward option %s at line %d does not match any bridge definition\n", buf, line);
            exit(1);
          }
          break;
        case 5:
          if (!set_anyport(buf)) {
            fprintf(stderr,"ERROR: anyport option %s at line %d does not match any bridge definition\n", buf, line);
            exit(1);
          }
          break;
        default:
          fprintf(stderr,"ERROR: weird state at line %d\n",line);
          exit(1);
        }
      }
    }
  }
  fclose(f);
}


int is_ethertype(struct DATA *d, int type)
{
  return ((d->data[13] == (type & 0xff)) &&
          (d->data[12] == (type >> 8)));
}


int is_decnet(struct DATA *data)
{
  return is_ethertype(data, ETHERTYPE_DECNET);
}


int is_lat(struct DATA *data)
{
  return (is_ethertype(data, ETHERTYPE_LAT) ||
          is_ethertype(data, ETHERTYPE_MOPDL) ||
          is_ethertype(data, ETHERTYPE_MOPRC));
}


unsigned long timedelta(struct timeval old)
{
  struct timeval now;
  unsigned long delta;
  gettimeofday(&now, NULL);
  delta = now.tv_sec - old.tv_sec;
  delta *= 1000000;
  delta += (now.tv_usec - old.tv_usec);
  return delta;
}


void throttle(int index)
{
  long delta;

  delta = timedelta(bridge[index].lasttime);
  bridge[index].throttle <<= 1;
  bridge[index].throttle += (delta < THROTTLETIME ? 1 : 0);

  if ((bridge[index].throttle & THROTTLEMASK) == THROTTLEMASK) {
    bridge[index].throttlecount++;
    usleep(THROTTLEDELAY);
    if (debug) fprintf(stdout,"Throttle active for bridge %d %d(%03o)\n",
                       index, bridge[index].throttlecount, bridge[index].throttle & 255);
  }
  gettimeofday(&bridge[index].lasttime,NULL);
}


int active(int index)
{
  if (bridge[index].passive == 0) return 1;
  if (timedelta(bridge[index].lastrcv) < PASSIVE_TMO) return 1;
  return 0;
}


/* do the actual sending */
void send_packet(int index, struct DATA *d)
{
  struct sockaddr_in sa;

  if (index == d->source) return; /* Avoid loopback of data. */
  if (bridge[index].types[d->type] == 0) return; /* Avoid sending unwanted frames */

  if ((((bridge[d->source].addr == 0) && (bridge[index].addr == 0))
    || ((bridge[d->source].addr != 0) && (bridge[index].addr != 0))) 
    && ((bridge[d->source].forward == 0) && (bridge[index].forward == 0))) return; /* Avoid forward */

  if (active(index)) {
    bridge[index].tcount++;
    throttle(index);
    if (bridge[index].addr == 0) {
      write(bridge[index].fd,d->data,d->len); /* Local network. */
    } else {
      sa.sin_family = AF_INET;        /* Remote network. */
      sa.sin_port = bridge[index].port;
      sa.sin_addr.s_addr = bridge[index].addr;
      if (sendto(bridge[index].fd,d->data,d->len,0,(struct sockaddr *)&sa,sizeof(sa)) == -1)
        perror("sendto");
    }
    bridge[index].lastptr = (bridge[index].lastptr+1) & 7;
    memcpy(bridge[index].last[bridge[index].lastptr],d->data,14);
  }
}


void register_source(struct DATA *d)
{
  unsigned short hash;
  struct HOST *h;

  hash = *(unsigned short *)(d->data+10);
  h = hosts[hash];
  while (h) {
    if (memcmp(h->mac, d->data+6, 6) == 0) {
      h->bridge = d->source;
      if(debug) { 
        gettimeofday(&h->lastseen, NULL);
        fprintf(stdout,"Got known host frame from bridge %d ", h->bridge);
        if (is_decnet(d)) {
          fprintf(stdout,"(%d.%d)\n", h->mac[5] >> 2, ((h->mac[5] & 3) << 8) + h->mac[4]);
        } else {
          fprintf(stdout, "(LAT or MOP data)\n");
        }
      }
      return;
    }
    h = h->next;
  }
  h = malloc(sizeof(struct HOST));
  h->next = hosts[hash];
  memcpy(h->mac,d->data+6,6);
  h->bridge = d->source;
  if(debug) {
    gettimeofday(&h->lastseen, NULL);
    fprintf(stdout,"Adding new hash entry for bridge %d ", h->bridge);
    if (is_decnet(d)) {
      fprintf(stdout,"(%d.%d)\n", h->mac[5] >> 2, ((h->mac[5] & 3) << 8) + h->mac[4]);
    } else {
      fprintf(stdout, "(LAT or MOP data)\n");
    }
  }
  hosts[hash] = h;
}


int locate_dest(struct DATA *d)
{
  unsigned short hash;
  struct HOST *h;

  if (d->data[0] & 1) return -1; /* Ethernet multicast */

  hash = *(unsigned short *)(d->data+4);
  h = hosts[hash];
  while (h) {
    if (memcmp(h->mac, d->data, 6) == 0) return h->bridge;
    h = h->next;
  }
  return -1;
}


void process_packet(struct DATA *d)
{
  int dst;
  int i;

  bridge[d->source].rcount++;
  gettimeofday(&bridge[d->source].lastrcv, NULL);
  for (i=0; i<8; i++) {
    if (memcmp(bridge[d->source].last[i],d->data,14) == 0) {
      return;
    }
  }

  if (is_decnet(d)) d->type = DECnet;
  if (is_lat(d)) d->type = LAT;

  if (bridge[d->source].types[d->type] == 0) return;
  if (d->type == -1) return;

  bridge[d->source].xcount++;

  register_source(d);
  dst = locate_dest(d);
  if (dst == -1) {
    int i;
    for (i=0; i<bcnt; i++) send_packet(i, d);
  } else {
    send_packet(dst, d);
  }
}


void dump_config()
{
  int i;
  struct in_addr addr = {0};

  fprintf(stdout,"Host table:\n");
  for (i=0; i<bcnt; i++) {
    addr.s_addr = bridge[i].addr;
    fprintf(stdout,"%d: %s %s:%d (Rx: %d Tx: %d (Drop rx: %d)) Active: %d Throttle: %d(%03o) Fwd: %d AnyP: %d\n",
           i,
           bridge[i].name,
           inet_ntoa(addr),
           ntohs(bridge[i].port),
           bridge[i].rcount,
           bridge[i].tcount,
           bridge[i].rcount - bridge[i].xcount,
           active(i),
           bridge[i].throttlecount,
           bridge[i].throttle & 255,
           bridge[i].forward,
           bridge[i].anyport);
  }
}


void dump_table()
{
  int i;
  time_t lastseen;

  fprintf(stdout,"Hash of known destinations:\n");
  for (i=0; i<HOST_HASH; i++) {
    struct HOST *h;
    h=hosts[i];
    while (h) {
      fprintf(stdout,"%02x:%02x:%02x:%02x:%02x:%02x -> %d",
             (unsigned char)h->mac[0],
             (unsigned char)h->mac[1],
             (unsigned char)h->mac[2],
             (unsigned char)h->mac[3],
             (unsigned char)h->mac[4],
             (unsigned char)h->mac[5],
             h->bridge);
      if ((unsigned char)h->mac[0] == 0xaa &&
          (unsigned char)h->mac[1] == 0x00 &&
          (unsigned char)h->mac[2] == 0x04 &&
          (unsigned char)h->mac[3] == 0x00) {
           fprintf(stdout," (%d.%d)", h->mac[5] >> 2, ((h->mac[5] & 3) << 8) + h->mac[4]);
      }
      lastseen = (time_t)h->lastseen.tv_sec; /* time_t != long */
      fprintf(stdout," last seen %s",ctime(&lastseen));  /* ctime() provides new line character */
      h = h->next;
    }
  }
}


void refresh_ipaddr()
{
  struct hostent *he;
  struct in_addr addr = {0};
  int i, found = 0;
  char* actual_addr, * new_addr;

  if (debug) dump_table();  /* placed here for convenience but has nothing in common with DNS lookups */
  if (debug) fprintf(stdout,"DECnet bridge now looking up hostnames\n");
  for (i=0; i<bcnt; i++) {
    addr.s_addr = bridge[i].addr;
    actual_addr = (char *)inet_ntoa(addr); 
    if ((strcmp(bridge[i].host,actual_addr) != 0)                     /* Avoid resolving dotted decimal addresses */ 
    && (ntohs(bridge[i].port) != 0)) {                                /* Avoid resolving local bridge definitions */
      if (timedelta(bridge[i].lastrcv) > (refresh_interval/2)) {
        if (debug) fprintf(stdout,"Resolving host %s for bridge %s ... ",bridge[i].host,bridge[i].name);
        if ((he = gethostbyname(bridge[i].host)) != NULL) {
          addr.s_addr = *(in_addr_t *)he->h_addr;
          found = -1;
        } else {
          found = inet_aton(bridge[i].host,&addr);
        }
        if (found) bridge[i].addr = addr.s_addr;
        if (debug) {
          if (found) {
            new_addr = (char *)inet_ntoa(addr);
            fprintf(stdout,"OK (old %s - new %s)\n", actual_addr, new_addr); 
          } else { 
            fprintf(stdout,"ERROR (old %s)\n", actual_addr);
          }
        }
      }
    }    
  }
}


void print_usage(char* app_name)
{
  fprintf(stdout, "Usage: %s [OPTION]...\n\n", app_name);
  fprintf(stdout, " -b ADDR     IP address to bind to (default 0.0.0.0 i.e. any address)\n");
  fprintf(stdout, " -c FILE     Non-default configuration file\n");
  fprintf(stdout, " -d          Enable debug\n");
  fprintf(stdout, " -h          Display this help and exit\n");
  fprintf(stdout, " -r TIME     Enable DNS cyclic lookup for remote host names every TIME seconds\n");
  fprintf(stdout, "             DNS lookup occurs only for bridges which did not transmit anything\n");
  fprintf(stdout, "             in the last (TIME / 2) seconds. Lookup disabled if not specified\n");
  fprintf(stdout, " -D          Run as a daemon\n");
  fprintf(stdout, " -P PORT     Non-default port used to receive frames from remote bridges\n");
  fprintf(stdout, "\n"); 
}


int main(int argc, char **argv)
{
  struct sockaddr_in sa,rsa;
  int i, hsock, pid, temp, parse_port, cmd_err = 0, daemon = 0;
  fd_set fds;
  socklen_t ilen;
  struct DATA d;
  char buf[8192];

  struct timeval refresh_time;
  struct in_addr bind_addr = {0};

  sa.sin_port = htons(DEFAULT_PORT);
  sa.sin_addr.s_addr = INADDR_ANY;

  if ((sd = socket(PF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("socket");
    exit(1);
  }

  fprintf(stdout, "\n  This is DECnet Bridge Version V%d.%d-%d\n\n", BRIDGE_VERSION, BRIDGE_RELEASE, BRIDGE_EDITLVL);

/* command line parsing and checking */
  if (1 < argc && argc < 12) {

    for (temp = 1; temp < argc; temp++) {
     
      if (!strcmp(argv[temp],"-h")) {
        cmd_err = 1;
  
      } else if (!strcmp(argv[temp],"-P")) {
        temp++;
        if (temp >= argc) {
          fprintf(stderr,"ERROR parsing command line: no port number given for -P\n");
          cmd_err = 1;
        } else if (!(strspn(argv[temp],"0123456789") == strlen(argv[temp]))) {
          fprintf(stderr,"ERROR parsing command line: value for -P is not numeric\n");
          cmd_err = 1;
        } else {
          parse_port = atoi(argv[temp]);
          if (parse_port < 1025 || parse_port > 32767) {
            fprintf(stderr,"ERROR parsing command line: value for -P is out of range (1025-32767)\n");
            cmd_err = 1;
          } else sa.sin_port = htons(parse_port);
        }
  
      } else if (!strcmp(argv[temp],"-d")) {
        debug = 1;
  
      } else if (!strcmp(argv[temp],"-r")) {
        temp++;
        if (temp >= argc) {
          fprintf(stderr,"ERROR parsing command line: no seconds given for -r\n");
          cmd_err = 1;
        } else if (!(strspn(argv[temp],"0123456789") == strlen(argv[temp]))) {
          fprintf(stderr,"ERROR parsing command line: value for -r is not numeric\n");
          cmd_err = 1;
        } else {
          refresh_interval = (atol(argv[temp]) * 1000000);
          if (refresh_interval < 180000000) {
            fprintf(stderr,"ERROR parsing command line: value for -r is too small (less than 180 seconds)\n");
            cmd_err = 1;
          } 
        } 
  
      } else if (!strcmp(argv[temp],"-c")) {
        temp++;
        if (temp >= argc) {
          fprintf(stderr,"ERROR parsing command line: no file name given for -c\n");
          cmd_err = 1;
        } else config_name=argv[temp];
  
      } else if (!strcmp(argv[temp],"-D")) {
        daemon = 1;
  
      } else if (!strcmp(argv[temp],"-b")) {
        temp++;
        if (temp >= argc) {
          fprintf(stderr,"ERROR parsing command line: no IP address given for -b\n");
          cmd_err = 1;
        } else if (!inet_aton(argv[temp],&bind_addr)) { 
          fprintf(stderr,"ERROR parsing command line: value for -b is not a valid IP address\n");
          cmd_err = 1;
        } else sa.sin_addr.s_addr = bind_addr.s_addr;

      } else {
        fprintf(stderr,"ERROR parsing command line: unknown parameter %s\n",argv[temp]);
        cmd_err = 1;
      }

    } /* end for */

  } else {
    cmd_err = 1;
  }

  if (cmd_err) {
    print_usage(argv[0]);
    exit(1);
  }

  sa.sin_family = AF_INET;
  if (bind(sd, (struct sockaddr*)&sa, sizeof(sa)) == -1) {
    perror("bind");
    exit(1);
  }

  read_conf();
  gettimeofday(&refresh_time, NULL);

  if (debug) dump_config();

  if (daemon) {
    pid = fork();
    if (pid == -1) {
      perror("fork");     // fork failed
      exit(1);
    } else if (pid == 0) {  // this is the child 
      setsid();
    } else {              // this is the parent
      _exit(0);
    }
  }

  while(1) {

    if (refresh_interval > 0) {
      if (timedelta(refresh_time) > refresh_interval) {
        refresh_ipaddr();
        gettimeofday(&refresh_time, NULL);
      }
    }

    FD_ZERO(&fds);
    hsock = 0;
    for (i=0; i<bcnt; i++) {
      FD_SET(bridge[i].fd, &fds);
      if (bridge[i].fd > hsock) hsock = bridge[i].fd;
    }
    if (select(hsock+1,&fds,NULL,NULL,NULL) == -1) {
      if (errno != EINTR) {
        perror("select");
        exit(1);
      }
    }

    for (i=0; i<bcnt; i++) {
      if (FD_ISSET(bridge[i].fd,&fds)) {
        d.source = i;
        d.type = -1;
        if (bridge[i].addr == 0) {
          struct pcap_pkthdr h;
          d.data = (char *)pcap_next(bridge[i].pcap, &h);
          d.len = h.caplen;
          if (d.data) {
            process_packet(&d);
          }
        } else {
          ilen = sizeof(rsa);
          if ((d.len = recvfrom(bridge[i].fd, buf, 1518, 0,
                                (struct sockaddr *)&rsa, &ilen)) > 0) {
            if ((d.source = lookup(&rsa)) >= 0) {
              d.data = buf;
              process_packet(&d);
            }
          }
        }
        FD_CLR(bridge[i].fd, &fds);
      }
    }
  }
}
